//
//	DataPG.h
//

#ifndef DATAPG_H
#define DATAPG_H

//#ifndef ENTRYPG_H
#include "EntryPG.h"
//#endif

//#ifndef MORPHPG_H
#include "MorphPG.h"
//#endif

//#pragma mark TexturePG

//
//	TexturePG class
//

class TexturePG
{
	// member objects
	public:
	
	enum RenderMode
	{
		kInvisable,
		kWireframe,
		kPolygon,
		kTexture
	};
	
	// member functions
	public:
	
	TexturePG(void);
	~TexturePG(void);
	
	void AddVertex(VertexPG*);
	void SetColor(unsigned char);
	void SetTexture(short w, short h);
	void SetMode(TexturePG::RenderMode);
	
	// data members
	public:
	
	VertexPG*			vertices[5];
	short				edgeCount;
	RenderMode			mode;
	unsigned char		*texture;
	unsigned char		color;
	short				width;
	short				height;
};

//#pragma mark TrianglePG

//
//	TrianglePG class
//

class TrianglePG
{
	//	member functions
	public:
	
	TrianglePG(void);
	void AddVertex(VertexPG*);
	void AddTexture(TexturePG*);
	
	//	data members
	public:
	VertexPG*			vertices[5];
	VertexPG			clipped[12];
	TexturePG*			texture;
	short				edgeCount;
	short				clipCount;
	short				xMin, yMin, zMin;
	short				xMax, yMax, zMax;
	bool				swapped;
};

//#pragma mark MeshPG

//
//	MeshPG class
//

class MeshPG : public MorphPG
{
	// member functions
	public:
	
	MeshPG(void);
	~MeshPG(void);
	virtual void Morph(void);
	
	void AddVertex(VertexPG*, long);
	void AddTriangle(TrianglePG*, long);
	void AddTexture(TexturePG*, long);
	
	private:
	
	void ClearVertices(void);
	void ClearTriangles(void);
	void ClearTextures(void);
	
	
	// data members:
	public:
	
	Applicant<VertexPG>*		vertices;
	Applicant<TrianglePG>*		triangles;
	Applicant<TexturePG>*		textures;
};

//#pragma mark LatticePG

//
//	LatticePG class
//

class LatticePG : public MorphPG
{
	//	member functions
	public:
	
	LatticePG(void);
	~LatticePG(void);
	
	void AddMesh(MeshPG*, char*);
	void RemoveMesh(char*);
	MeshPG* FindMesh(char*);
	void LinkMesh(char* super, char* sub);
	void Morph(void);
	
	void Bounds(void);
	
	//	data members
	public:
	
	Frame<MeshPG>*			meshes;
	short					left;
	short					top;
	short					right;
	short					bottom;
};

//#pragma mark DataPG

//
//	DataPG class
//

class DataPG : public MorphPG
{
	// member functions
	public:
	
	DataPG(void);
	~DataPG(void);
	void ClearLattices(void);
	void ClearTriangles(void);
	
	void Rotate(short, short, short);
	void Scale(float, float, float);
	void Translate(short, short, short);
	
	LatticePG*	CreateLattice(char* id);
	void DeleteLattice(char*);
	LatticePG* FindLattice(char*);
	MeshPG* FindMesh(char* lattice, char* mesh);
	void RegisterTriangle(TrianglePG*);
	void RemoveTriangle(TrianglePG*);
	
	// data members
	public:
	
	Entry<TrianglePG>*		triangles;
	Frame<LatticePG>*		lattices;
};

#endif

//
//	DataPG.h
//