//
//	Entry.h
//

#ifndef ENTRYPG_H
#define	ENTRYPG_H

//#ifndef _STRING_H_
#include <string.h>
//#endif

//
//	Entry Template class
//

template <class obj> class Entry
{
	//	member functions
	public:
	
	Entry(obj*);
	~Entry(void);
	
	void AddEntry(Entry<obj>*);
	void RemoveEntry(Entry<obj>*);
	Entry<obj>*	Find(obj*);
	Entry<obj>*	FirstEntry(void);
	Entry<obj>* LastEntry(void);
	
	// data members
	public:
	
	Entry<obj>*		prior;
	Entry<obj>*		next;
	obj*			content;
};

//#pragma mark Applicant

//
//	Applicant template class
//

template <class obj> class Applicant : public Entry<obj>
{
	//	member functions
	public:
	
	Applicant(obj*, long);
	Applicant<obj>* Find(long);
	
	//	data members
	public:
	
	long	moniker;
};

//#pragma mark Frame

//
//	Frame template class
//

template <class obj> class Frame : public Entry<obj>
{
	//	member functions
	public:
	
	Frame(obj*, char*);
	Frame<obj>* Find(char*);
	void AddEntry(Frame<obj>*);
	
	//	data members
	public:
	
	char	moniker[80];
};

//#pragma mark -

//
//	Entry::Entry()
//
//	constuctor
//

template <class obj> Entry<obj>::Entry(obj*  gizmo)
{
	prior = nil;
	next = nil;
	content = gizmo;
}

//
//	Entry::~Entry()
//
//	destructor
//

template <class obj> Entry<obj>::~Entry(void)
{
	RemoveEntry(this);
}

//
//	Entry::AddEntry()
//

template <class obj> void Entry<obj>::AddEntry(Entry<obj>* gizmo)
{
	gizmo->prior = this;
	gizmo->next = next;
	
	if(next != nil)
		next->prior = gizmo;
		
	next = gizmo;
}

//
//	Entry::RemoveEntry()
//

template <class obj> void Entry<obj>::RemoveEntry(Entry<obj>* gizmo)
{
	Entry<obj>*		pre;
	Entry<obj>*		post;
	
	pre = gizmo->prior;
	post = gizmo->next;
	
	gizmo->prior = nil;
	gizmo->next = nil;
	
	if(pre != nil)
		pre->next = post;
	
	if(post != nil)
		post->prior = pre;
}

//
//	Entry::Find()
//

template <class obj> Entry<obj>* Entry<obj>::Find(obj* gizmo)
{
	Entry<obj>*		current = FirstEntry();
	
	while(current != nil)
	{
		Entry<obj>*	nextor;
		nextor = current->next;
		
		if(current->content == gizmo)
			return current;
		
		current = nextor;
	}
	return current;
}

//
//	Entry::FirstEntry()
//

template <class obj> Entry<obj>* Entry<obj>::FirstEntry(void)
{
	Entry<obj>*		current;
	
	current = this;
	while(current != nil)
	{
		Entry<obj>*	prioron;
		prioron = current->prior;
		
		if(prioron == nil)
			return current;
			
		current = prioron;
	}
	return current;
}

//
//	Entry::LastEntry()
//

template <class obj> Entry<obj>* Entry<obj>::LastEntry(void)
{
	Entry<obj>*		current;
	
	current = this;
	while(current != nil)
	{
		Entry<obj>*	nextor;
		nextor = current->next;
		
		if(nextor == nil)
			return current;
		
		current = nextor;
	}
	return current;
}

//#pragma mark -

//
//	Applicant::Applicant()
//

template <class obj> Applicant<obj>::Applicant(obj* gizmo, long title) : Entry<obj>(gizmo)
{
	moniker = title;
}

//
//	Applicant::Find()
//

template <class obj> Applicant<obj>* Applicant<obj>::Find(long title)
{
	Applicant<obj>*		current;
	current = (Applicant<obj>*)FirstEntry();
	
	while(current != nil)
	{
		Applicant<obj>*		nextor;
		nextor = (Applicant<obj>*)current->next;
		
		if(current->moniker == title)
			return current;
			
		current = nextor;
	}
	
	return nil;
}

//#pragma mark -

//
//	Frame::Frame()
//
//	constructor
//

template <class obj> Frame<obj>::Frame(obj* gizmo, char* title) : Entry<obj>(gizmo)
{
	memcpy(moniker, title, 80);
}

//
//	Frame::AddEntry()
//

template <class obj> void Frame<obj>::AddEntry(Frame<obj>* gizmo)
{
	Frame<obj>*		last;
	last = (Frame<obj>*)LastEntry();
	
	last->next = gizmo;
	gizmo->prior = last;
}

//
//	Frame::Find()
//

template <class obj> Frame<obj>* Frame<obj>::Find(char* title)
{
	Frame<obj>*		current;
	current = (Frame<obj>*)FirstEntry();
	
	while(current != nil)
	{
		Frame<obj>*		nextor;
		nextor = (Frame<obj>*)current->next;
		
		if(strcmp(current->moniker, title) == 0)
			return current;
			
		current = nextor;
	}
	
	return nil;
}

#endif

//
//	Entry.cp
//