//
//	FilePG.cp
//

#define inMetrowerks2
#ifndef inMetrowerks2
typedef unsigned char bool;
#endif

//#define inMacOS
#ifndef inMacOS
#define nil 0L
#define kEndLine '\n'
#define kDirectory "/"
#else
#define kEndLine '\n'
#define kDirectory ":"
#endif

#define FILEPG_CP

#include <fstream.h>
#include "EntryPG.h"
#include "MorphPG.h"
#include "DataPG.h"
#include "FilePG.h"


//#pragma mark -

//
//	FilePG::FilePG()
//
//	constructor
//

FilePG::FilePG(void) : DataPG()
{
	fin = nil;
	fLattice = nil;
	fMesh = nil;
	fTexture = nil;
}

//
//	FilePG::~FilePG()
//
//	destructor
//

FilePG::~FilePG(void)
{
	//	nothing yet
}

//
//	FilePG::Load()
//

bool FilePG::Load(char* fileName)
{
	fin = new fstream(fileName, ios::in);
	
	if(fin->bad())
		return false;
	
	while(fin->eof() == false)
	{
		fin->getline(fLine, 80, ' ');
		Token();
	}
	
	return true;
}

//
//	FilePG::Token()
//

void FilePG::Token(void)
{
	if(strcmp(fLine, "End:") == 0)
		End();
	else if(strcmp(fLine, "Rotate:") == 0)
		DoRotate();
	else if(strcmp(fLine, "Scale:") == 0)
		DoScale();
	else if(strcmp(fLine, "Translate:") == 0)
		DoTranslate();
	else if(strcmp(fLine, "Lattice:") == 0)
		FileCreateLattice();
	else if(strcmp(fLine, "Mesh:") == 0)
		CreateMesh();
	else if(strcmp(fLine, "Origin:") == 0)
		SetOrigin();
	else if(strcmp(fLine, "Owner:") == 0)
		SetOwner();
	else if(strcmp(fLine, "Vertex:") == 0)
		CreateVertex();
	else if(strcmp(fLine, "Triangle:") == 0)
		CreateTriangle();
	else if(strcmp(fLine, "Texture:") == 0)
		CreateTexture();
	else if(strcmp(fLine, "Point:") == 0)
		SetPoint();
	else if(strcmp(fLine, "Width:") == 0)
		SetWidth();
	else if(strcmp(fLine, "Height:") == 0)
		SetHeight();
	else if(strcmp(fLine, "Color:") == 0)
		SetColor();
	else if(strcmp(fLine, "Mode:") == 0)
		SetMode();
	else if(strcmp(fLine, "Prefix:") == 0)
		TexturePrefix();
	else
		fin->getline(fLine, 80, kEndLine);
}

//
//	FilePG::End()
//

void FilePG::End(void)
{
	fin->getline(fLine, 80, kEndLine);
	
	if(strcmp(fLine, "Lattice") == 0)
		fLattice = nil;
	else if(strcmp(fLine, "Mesh") == 0)
		fMesh = nil;
	else if(strcmp(fLine, "Texture") == 0)
		fTexture = nil;
}

//#pragma mark -

//
//	FilePG::FileCreateLattice()
//
//	uses persistant data
//

void FilePG::FileCreateLattice(void)
{
	char		id[80];
	fin->getline(id, 80, kEndLine);
	
	fLattice = CreateLattice(id);
}

//
//	FilePG::CreateMesh()
//
//	uses persistant data
//

void FilePG::CreateMesh(void)
{
	char		id[80];
	fin->getline(id, 80, kEndLine);
	
	if(fLattice == nil)
		return;
	
	fMesh = new MeshPG();
	fLattice->AddMesh(fMesh, id);
}

//
//	FilePG::SetOrigin()
//
//	uses persistant data
//

void FilePG::SetOrigin(void)
{
	short	x;
	short 	y;
	short	z;
	
	*fin >> x;	fin->getline(fLine, 80, ' ');
	*fin >> y;	fin->getline(fLine, 80, ' ');
	*fin >> z;	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh != nil)
		fMesh->Origin(x, y, z);
	else if(fLattice != nil)
		fLattice->Origin(x, y, z);
}

//
//	FilePG::SetOwner()
//

void FilePG::SetOwner(void)
{
	char		id[80];
	MeshPG*		super;
	fin->getline(id, 80, kEndLine);
	
	if((fLattice == nil) || (fMesh == nil))
		return;
		
	super = fLattice->FindMesh(id);
	if(super == nil)
		return;
		
	fMesh->Owner(super);
}

//
//	FilePG::CreateVertex()
//

void FilePG::CreateVertex(void)
{
	long		id;
	short		x;
	short		y;
	short		z;
	VertexPG*	fVertex;
	
	*fin >> id;	fin->getline(fLine, 80, ' ');
	*fin >> x;	fin->getline(fLine, 80, ' ');
	*fin >> y;	fin->getline(fLine, 80, ' ');
	*fin >> z;	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh == nil)
		return;
	
	fVertex = new VertexPG(x, y, z);
	fMesh->AddVertex(fVertex, id);
	
	fVertex = nil;
}

//
//	FilePG::CreateTriangle()
//

void FilePG::CreateTriangle(void)
{
	long			id;
	char			dummy;
	TrianglePG*		fTriangle;
	
	if((fMesh == nil) || (fLattice == nil))
	{
		fin->getline(fLine, 80, kEndLine);
		return;
	}
	
	*fin >> id;
	fTriangle = new TrianglePG();
	fMesh->AddTriangle(fTriangle, id);
	RegisterTriangle(fTriangle);
	
	fin->get(dummy);
	
	while(dummy != kEndLine)
	{
		VertexPG*		v;
		*fin >> id;
		
		fin->get(dummy);
		if(dummy == kEndLine)
		{
			TexturePG*	t = (fMesh->textures->Find(id))->content;
			fTriangle->AddTexture(t);
			break;
		}
		
		v = (fMesh->vertices->Find(id))->content;
		fTriangle->AddVertex(v);
	}
}

//#pragma mark -

//
//	FilePG::CreateTexture()
//
//	uses persistant data
//

void FilePG::CreateTexture(void)
{
	short	id;
	*fin >> id;
	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh == nil)
		return;
	
	fTexture = new TexturePG();
	fMesh->AddTexture(fTexture, id);
}	

//
//	FilePG::SetPoint()
//
//	uses persistant data
//

void FilePG::SetPoint(void)
{
	short	x;
	short	y;
	
	*fin >> x;		fin->getline(fLine, 80, ' ');
	*fin >> y;		fin->getline(fLine, 80, kEndLine);
	
	if(fTexture == nil)
		return;
		
	fTexture->AddVertex(new VertexPG(x, y));
}

//
//	FilePG::SetWidth()
//
//	uses persistant data
//

void FilePG::SetWidth(void)
{
	short	w;
	*fin >> w;
	fin->getline(fLine, 80, kEndLine);
	
	if(fTexture != nil)
	{
		fTexture->width = w;
	}
}

//
//	FilePG::SetHeight()
//
//	uses persistant data
//

void FilePG::SetHeight(void)
{
	short	h;
	*fin >> h;
	fin->getline(fLine, 80, kEndLine);
	
	if(fTexture != nil)
	{
		fTexture->height = h;
	}
}

//
//	FilePG::SetColor()
//
//	uses persistant data
//

void FilePG::SetColor(void)
{
	short	c;
	*fin >> c;
	fin->getline(fLine, 80, kEndLine);
	
	if(fTexture != nil)
	{
		fTexture->color = (unsigned char)c;
	}
}

//
//	FilePG::SetMode()
//
//	uses persistant data
//

void FilePG::SetMode(void)
{
	fin->getline(fLine, 80, kEndLine);
	
	if(fTexture == nil)
		return;
	
	if(strcmp(fLine, "kWireframe") == 0)
		fTexture->SetMode(TexturePG::kWireframe);
	else if(strcmp(fLine, "kPolygon") == 0)
		fTexture->SetMode(TexturePG::kPolygon);
	else if(strcmp(fLine, "kTexture") == 0)
		fTexture->SetMode(TexturePG::kTexture);
	else
		fTexture->SetMode(TexturePG::kInvisable);
}	

//
//	FilePG::TexturePrefix()
//

void FilePG::TexturePrefix(void)
{
	char		id[80];
	fin->getline(id, 80, kEndLine);
	
	if(fTexture == nil)
		return;
	
	fstream		prefix;
	prefix.open(id, ios::in | ios::binary);
	if(prefix.bad())
		return;
	
	unsigned long	w = fTexture->width;
	unsigned long	h = fTexture->height;
	unsigned long	size = (w * h);
	
	fTexture->texture = new unsigned char[size];
	for(unsigned long i = 0; i < size; i++)
	{
		fTexture->texture[i] = prefix.get();
		if(prefix.eof())
			return;
	}
	prefix.close();
}

//#pragma mark -

//
//	FilePG::DoRotate()
//

void FilePG::DoRotate(void)
{
	short	x;
	short 	y;
	short	z;
	
	*fin >> x;	fin->getline(fLine, 80, ' ');
	*fin >> y;	fin->getline(fLine, 80, ' ');
	*fin >> z;	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh != nil)
		fMesh->Rotate(x, y, z);
	else if(fLattice != nil)
		fLattice->Rotate(x, y, z);
	else
		Rotate(x, y, z);
}

//
//	FilePG::DoScale()
//

void FilePG::DoScale(void)
{
	float	x;
	float 	y;
	float	z;
	
	*fin >> x;	fin->getline(fLine, 80, ' ');
	*fin >> y;	fin->getline(fLine, 80, ' ');
	*fin >> z;	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh != nil)
		fMesh->Scale(x, y, z);
	else if(fLattice != nil)
		fLattice->Scale(x, y, z);
}

//
//	FilePG::DoTranslate()
//

void FilePG::DoTranslate(void)
{
	short	x;
	short 	y;
	short	z;
	
	*fin >> x;	fin->getline(fLine, 80, ' ');
	*fin >> y;	fin->getline(fLine, 80, ' ');
	*fin >> z;	fin->getline(fLine, 80, kEndLine);
	
	if(fMesh != nil)
		fMesh->Translate(x, y, z);
	else if(fLattice != nil)
		fLattice->Translate(x, y, z);
	else
		Translate(x, y, z);
}

//
//	FilePG.cp
//