//
//	MorphPG.cp
//

#define inMetrowerks2
#ifndef inMetrowerks2
typedef unsigned char bool;
#endif

//#define inMacOS
#ifndef inMacOS
#define nil 0L
#endif

#include "MorphPG.h"

//#pragma mark -

//
//	VertexPG::VertexPG()
//
//	constructor
//

VertexPG::VertexPG(void)
{
	x = lx = wx = 0;
	y = ly = wy = 0;
	z = lz = wz = 0;
}

//
//	VertexPG::VertexPG()
//
//	constructor
//

VertexPG::VertexPG(short ix, short iy)
{
	sx = ix;
	sy = iy;
}

//
//	VertexPG::VertexPG()
//
//	constructor
//

VertexPG::VertexPG(short ix, short iy, short iz)
{
	x = lx = wx = ix;
	y = ly = wy = iy;
	z = lz = wz = iz;
}

//#pragma mark -

//
//	MorphPG::MorphPG()
//
//	constructor
//

MorphPG::MorphPG(void)
{
	float	radians;
	float	sinResult, cosResult;
	
	origin = new VertexPG(0,0);
	owner = nil;
	
	for(short i = 0; i < 360; i++)
	{
		radians = i * 0.017453292;
		
		sinResult = sin(radians);
		cosResult = cos(radians);
		
		sinTable[i] = (long int)(sinResult * 256);
		cosTable[i] = (long int)(cosResult * 256);
	}
	
	Reset();
}

//
//	MorphPG::~MorphPG()
//
//	destructor

MorphPG::~MorphPG(void)
{
	delete origin;
}

//
//	MorphPG::Origin()
//

void MorphPG::Origin(short x, short y, short z)
{
	origin->x = origin->lx = origin->wx = x;
	origin->y = origin->ly = origin->wy = y;
	origin->z = origin->lz = origin->wz = z;
}

//
//	MorphPG::Owner()
//

void MorphPG::Owner(MorphPG* super)
{
	owner = super;
}

//
//	MorphPG::Sever()
//

void MorphPG::Sever(void)
{
	owner = nil;
}

//
//	MorphPG::Transform()
//

void MorphPG::Transform(VertexPG* ref)
{
	short		alpha, beta, gamma, delta;
	
	// multiply against the matrix
	alpha = (ref->lx * transformMatrix[0][0] +
			ref->ly * transformMatrix[1][0] +
			ref->lz * transformMatrix[2][0] +
			transformMatrix[3][0]) >> 8;
	
	beta = 	(ref->lx * transformMatrix[0][1] +
			ref->ly * transformMatrix[1][1] +
			ref->lz * transformMatrix[2][1] +
			transformMatrix[3][1]) >> 8;
	
	gamma = (ref->lx * transformMatrix[0][2] +
			ref->ly * transformMatrix[1][2] +
			ref->lz * transformMatrix[2][2] +
			transformMatrix[3][2]) >> 8;
	
	delta = (ref->lx * transformMatrix[0][3] +
			ref->ly * transformMatrix[1][3] +
			ref->lz * transformMatrix[2][3] +
			transformMatrix[3][3]) >> 8;
			
	ref->tx = alpha;
	ref->ty = beta;
	ref->tz = gamma;
	ref->tt = delta;
}		

//
//	MorphPG::Morph()
//
//	virtual
//

void MorphPG::Morph(void)
{
	FixedPG		tempMatrix[4][4];
	
	
	IdentityMatrix(transformMatrix);
	RotateMatrix();
	ScaleMatrix();
	
	if(owner != nil)
	{
		//	subtransform the origin
		owner->Transform(origin);
		origin->wx = owner->origin->wx + origin->tx;
		origin->wy = owner->origin->wy + origin->ty;
		origin->wz = owner->origin->wz + origin->tz;
		
		//	subtransform the matrix
		MatrixMultiply(	owner->transformMatrix,
						transformMatrix,
						tempMatrix	);
		
		CopyMatrix(tempMatrix, transformMatrix);
	}
}

//
//	MorphPG::Rotate()
//

void MorphPG::Rotate(short x, short y, short z)
{
	xAngle += x;
	yAngle += y;
	zAngle += z;
	
	//	check for angle wrap
	while(xAngle >= 360)	xAngle -= 360;
	while(xAngle < 0)		xAngle += 360;
	while(yAngle >= 360)	yAngle -= 360;
	while(yAngle < 0)		yAngle += 360;
	while(zAngle >= 360)	zAngle -= 360;
	while(zAngle < 0)		zAngle += 360;
}

//
//	MorphPG::Scale()
//

void MorphPG::Scale(float x, float y, float z)
{
	xScale *= x;
	yScale *= y;
	zScale *= z;
}

//
//	MorphPG::Translate()
//

void MorphPG::Translate(short x, short y, short z)
{
	origin->lx += x;
	origin->ly += y;
	origin->lz += z;
	
	//	handles the ownerless case
	origin->wx += x;
	origin->wy += y;
	origin->wz += z;
}

//
//	MorphPG::Reset()
//

void MorphPG::Reset(void)
{
	IdentityMatrix(transformMatrix);
	
	xAngle = 0;
	yAngle = 0;
	zAngle = 0;
	
	xScale = 1;
	yScale = 1;
	zScale = 1;
}

//#pragma mark -

//
//	MorphPG::RotateMatrix()
//

void MorphPG::RotateMatrix(void)
{
	FixedPG	rotateMatrix[4][4];
	FixedPG	tempMatrix1[4][4];
	FixedPG	tempMatrix2[4][4];
	FixedPG	xSin, xCos;
	FixedPG	ySin, yCos;
	FixedPG	zSin, zCos;
	
	//	set our FixedPG trig values
	xSin = Sin(xAngle);
	ySin = Sin(yAngle);
	zSin = Sin(zAngle);
	
	xCos = Cos(xAngle);
	yCos = Cos(yAngle);
	zCos = Cos(zAngle);
	
	// set the rotateMatrix to the y rotation
	rotateMatrix[0][0] = yCos;
	rotateMatrix[0][1] = 0;
	rotateMatrix[0][2] = -ySin;
	rotateMatrix[0][3] = 0;
	
	rotateMatrix[1][0] = 0;
	rotateMatrix[1][1] = 1 << 8;
	rotateMatrix[1][2] = 0;
	rotateMatrix[1][3] = 0;
	
	rotateMatrix[2][0] = ySin;
	rotateMatrix[2][1] = 0;
	rotateMatrix[2][2] = yCos;
	rotateMatrix[2][3] = 0;
	
	rotateMatrix[3][0] = 0;
	rotateMatrix[3][1] = 0;
	rotateMatrix[3][2] = 0;
	rotateMatrix[3][3] = 1 << 8;
	
	MatrixMultiply(transformMatrix, rotateMatrix, tempMatrix1);
	
	// set the rotateMatrix to the x rotation
	rotateMatrix[0][0] = 1 << 8;
	rotateMatrix[0][1] = 0;
	rotateMatrix[0][2] = 0;
	rotateMatrix[0][3] = 0;
	
	rotateMatrix[1][0] = 0;
	rotateMatrix[1][1] = xCos;
	rotateMatrix[1][2] = xSin;
	rotateMatrix[1][3] = 0;
	
	rotateMatrix[2][0] = 0;
	rotateMatrix[2][1] = -xSin;
	rotateMatrix[2][2] = xCos;
	rotateMatrix[2][3] = 0;
	
	rotateMatrix[3][0] = 0;
	rotateMatrix[3][1] = 0;
	rotateMatrix[3][2] = 0;
	rotateMatrix[3][3] = 1 << 8;
	
	MatrixMultiply(tempMatrix1, rotateMatrix, tempMatrix2);
	
	// set rotateMatrix to the z rotation
	rotateMatrix[0][0] = zCos;
	rotateMatrix[0][1] = zSin;
	rotateMatrix[0][2] = 0;
	rotateMatrix[0][3] = 0;
	
	rotateMatrix[1][0] = -zSin;
	rotateMatrix[1][1] = zCos;
	rotateMatrix[1][2] = 0;
	rotateMatrix[1][3] = 0;
	
	rotateMatrix[2][0] = 0;
	rotateMatrix[2][1] = 0;
	rotateMatrix[2][2] = 1 << 8;
	rotateMatrix[2][3] = 0;
	
	rotateMatrix[3][0] = 0;
	rotateMatrix[3][1] = 0;
	rotateMatrix[3][2] = 0;
	rotateMatrix[3][3] = 1 << 8;
	
	MatrixMultiply(tempMatrix2, rotateMatrix, tempMatrix1);
	
	// finish up
	CopyMatrix(tempMatrix1, transformMatrix);
}

//
//	MorphPG::ScaleMatrix()
//

void MorphPG::ScaleMatrix(void)
{
	FixedPG	scaleMatrix[4][4];
	FixedPG	xScalar;
	FixedPG	yScalar;
	FixedPG	zScalar;
	
	xScalar = (FixedPG)(xScale * 256);
	yScalar = (FixedPG)(yScale * 256);
	zScalar = (FixedPG)(zScale * 256);
	
	// init our scale matrix
	scaleMatrix[0][0] = xScalar;
	scaleMatrix[0][1] = 0;
	scaleMatrix[0][2] = 0;
	scaleMatrix[0][3] = 0;
	
	scaleMatrix[1][0] = 0;
	scaleMatrix[1][1] = yScalar;
	scaleMatrix[1][2] = 0;
	scaleMatrix[1][3] = 0;
	
	scaleMatrix[2][0] = 0;
	scaleMatrix[2][1] = 0;
	scaleMatrix[2][2] = zScalar;
	scaleMatrix[2][3] = 0;
	
	scaleMatrix[3][0] = 0;
	scaleMatrix[3][1] = 0;
	scaleMatrix[3][2] = 0;
	scaleMatrix[3][3] = 1 << 8;
	
	// scale our matrix and store it in
	// a temp veriable
	FixedPG		tempMatrix[4][4];
	
	MatrixMultiply(transformMatrix, scaleMatrix, tempMatrix);
	
	// finish the op by copying the tempMatrix to
	// the class transFormMatrix
	CopyMatrix(tempMatrix, transformMatrix);
}


//
//	MorphPG::IdentityMatrix()
//

void MorphPG::IdentityMatrix(FixedPG matrix[4][4])
{
	matrix[0][0] = 1 << 8;	matrix[0][1] = 0;
	matrix[0][2] = 0;		matrix[0][3] = 0;
	
	matrix[1][0] = 0;		matrix[1][1] = 1 << 8;
	matrix[1][2] = 0;		matrix[1][3] = 0;
	
	matrix[2][0] = 0;		matrix[2][1] = 0;
	matrix[2][2] = 1 << 8;	matrix[2][3] = 0;
	
	matrix[3][0] = 0;		matrix[3][1] = 0;
	matrix[3][2] = 0;		matrix[3][3] = 1 << 8;
}

//
//	MorphPG::CopyMatrix()
//

void MorphPG::CopyMatrix(	FixedPG sourceMatrix[4][4],
							FixedPG destMatrix[4][4]	)
{
	for(short i = 0; i < 4; i++)
	{
		destMatrix[i][0] = sourceMatrix[i][0];
		destMatrix[i][1] = sourceMatrix[i][1];
		destMatrix[i][2] = sourceMatrix[i][2];
		destMatrix[i][3] = sourceMatrix[i][3];
	}
}

//
//	MorphPG::MatrixMultiply()
//

void MorphPG::MatrixMultiply(	FixedPG matrix1[4][4],
								FixedPG matrix2[4][4],
								FixedPG final[4][4]	)
{
	FixedPG	temp;
	for(short i = 0; i < 4; i++)
	{
		for(short j = 0; j < 4; j++)
		{
			temp  = (matrix1[i][0] * matrix2[0][j]);
			temp += (matrix1[i][1] * matrix2[1][j]);
			temp += (matrix1[i][2] * matrix2[2][j]);
			temp += (matrix1[i][3] * matrix2[3][j]);
			
			temp = temp >> 8;
			final[i][j] = temp;
			
		}
	}
}

//#pragma mark -

//
//	MorphPG::Sin()
//

FixedPG MorphPG::Sin(short angle)
{
	return sinTable[angle];
}

//
//	MorphPG::Cos()
//

FixedPG MorphPG::Cos(short angle)
{
	return cosTable[angle];
}

//
//	MorphPG.cp
//