//
//	Mapper.cp
//

#include <View.h>
#include <Bitmap.h>
#include <TranslationUtils.h>
#include "RenderMap.h"

#pragma mark Mapper
#include "Mapper.h"

#pragma mark -

//
//	Mapper::Mapper()
//
//	constructor
//

Mapper::Mapper(BRect r) : BView(r, "Mapper", B_FOLLOW_NONE, B_WILL_DRAW)
{
	map = BTranslationUtils::GetBitmap("pgTexture.tiff");
}

//
//	Mapper::~Mapper()
//
//	destructor
//

Mapper::~Mapper(void)
{
	//	nothing yet
}

//
//	Mapper::SetImage()
//

void Mapper::SetImage(BMessage* message)
{
	entry_ref	ref;
	if(message->FindRef("refs", &ref) != B_OK)
		return;
		
	BEntry		entry(&ref);
	BPath		path(&entry);
	BBitmap*	newMap = BTranslationUtils::GetBitmapFile(path.Path());
	
	if(newMap != NULL)
	{
		if(map != NULL)
			delete map;
			
		map = new BBitmap(newMap->Bounds(), B_COLOR_8_BIT, true, true);
		BView*	v = new BView(newMap->Bounds(),"", B_FOLLOW_NONE, B_WILL_DRAW);
		map->AddChild(v);
		if(map->Lock())
		{
			v->DrawBitmap(newMap, BPoint(0,0));
			map->Unlock();
			delete newMap;
		}
		
		BRect	r;
		float	w;
		float	h;
		
		r = map->Bounds();
		w = r.Width();
		h = r.Height();
		
		w += 40;
		h += 40;
		ResizeTo(w, h);
		
		BMenuBar*	menuBar = (BMenuBar*)Window()->FindView("menuBar");
		h += menuBar->Bounds().Height() + 1;
		
		Window()->ResizeTo(w, h);
		ConvertToArray();
	}
}

//
//	Mapper::SaveImage()
//

void Mapper::SaveImage(BMessage* message)
{
	entry_ref	dir;
	if(message->FindRef("directory", &dir) != B_OK)
		return;
		
	BDirectory		bdir(&dir);
	const char*		name = message->FindString("name");
	if(name == NULL)
		return;
		
	BFile	file(&bdir, name, B_WRITE_ONLY | B_CREATE_FILE | B_ERASE_FILE);
	if(file.InitCheck() != B_OK)
	{
		BAlert*		alert;
		alert = new BAlert(NULL, "Could not create file.", "%$#@!");
		alert->Go();
		return;
	}
	
	file.Write(ConvertToArray(), ArraySize());
}

//
//	Mapper::ArraySize()
//

unsigned long Mapper::ArraySize(void)
{
	unsigned long	width = map->Bounds().Width();
	unsigned long	height = map->Bounds().Height();
	char*			baseAddress = (char*)map->Bits();
	return (width * height);
}

//
//	Mapper::ConvertToArray()
//

char* Mapper::ConvertToArray(void)
{
	char*			baseAddress = (char*)map->Bits();
	char*			rawData = new char[ArraySize()];
	unsigned long	rowBytes = map->BytesPerRow();
	short			width = map->Bounds().Width();
	short			height = map->Bounds().Height();
	
	for(short y = 0; y < height; y++)
	{
		for(short x = 0; x < width; x++)
		{
			unsigned long	index;
			unsigned long	bufferIndex;
			
			index = (rowBytes * y) + x;
			bufferIndex = (width * y) + x;
			
			rawData[bufferIndex] = baseAddress[index];
		}
	}
	
	return rawData;
}

//
//	Mapper::Draw()
//

void Mapper::Draw(BRect)
{
	if(Window()->Lock() == false)
		return;
	
	BRect	area;
	area = Bounds();
	
	SetHighColor(130, 130, 130);
	FillRect(area);
		
	SetDrawingMode(B_OP_OVER);
	if(map != NULL)
		DrawBitmap(map, BPoint(20,20));
	
	Window()->Unlock();
}

//
//	Mapper::AttachedToWindow()
//

void Mapper::AttachedToWindow(void)
{
	BRect	r;
	float	w;
	float	h;
	
	r = map->Bounds();
	w = r.Width();
	h = r.Height();
	
	w += 40;
	h += 40;
	ResizeTo(w, h);
	
	BMenuBar*	menuBar = (BMenuBar*)Window()->FindView("menuBar");
	h += menuBar->Bounds().Height() + 1;
	
	Window()->ResizeTo(w, h);
}

//
//	Mapper.cp
//