//
//	RenderPG.cp
//

#define inMetrowerks2
#ifndef inMetrowerks2
typedef unsigned char bool;
#endif

//#define inMacOS
#ifndef inMacOS
#define nil 0L
#endif

#include "EntryPG.h"
#include <View.h>

#include "MorphPG.h"
#include "DataPG.h"
#include "FilePG.h"

//#pragma mark RenderPG
#include "RenderPG.h"

#define kProjectConstant 200

//#pragma mark -

//
//	RenderPG::RenderPG()
//
//	constructor
//

RenderPG::RenderPG(BRect rect, char* name, BBitmap* map) : BView(rect, name, B_FOLLOW_NONE, B_WILL_DRAW)
{
	width = (short int)rect.Width();
	height = (short int)rect.Height();
	
	frameBuffer = map;
	rowBytes = (unsigned long)frameBuffer->BytesPerRow();
	baseAddress = (char*)frameBuffer->Bits();
}

//
//	RenderPG::~RenderPG()
//
//	destructor
//

RenderPG::~RenderPG(void)
{
	delete lattices;
}

//
//	RenderPG::MesssageReceived()
//

void RenderPG::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		default:
			BView::MessageReceived(message);
			break;
	}
}

//
//	RenderPG::Draw()
//

void RenderPG::Draw(BRect)
{
	Render();
}

//
//	RenderPG::Render()
//

void RenderPG::Render(void)
{	
	//	begin the render cycle
	rowBytes = (unsigned long)frameBuffer->BytesPerRow();
	baseAddress = (char*)frameBuffer->Bits();
	Erase();
	
	Entry<LatticePG>*		current;
	LatticePG*				latti;
	current = (Entry<LatticePG>*)lattices;
	
	MorphPG::Morph();
	
	while(current != nil)
	{
		Entry<LatticePG>*		next;
		next = current->next;
		latti = current->content;
		
		latti->Morph();
		Unimatrix(latti);
		ZClip(latti);
		Project(latti);

		if(next == nil)
			break;
			
		current = next;
	}
	
	XYClip();
	Extents();
	ZSort();
	Painters();
	DrawTriangles();
}

//#pragma mark -

//
//	RenderPG::Erase()
//

void RenderPG::Erase(void)
{
	long*	ptr;
	long	size;
	
	size = ((height + 1) * rowBytes) / 32;
	ptr = (long*)baseAddress;
	
	do
	{
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
		*ptr++ =0xFFFFFFFF;
	}
	while(--size);
}

//
//	RenderPG::Unimatrix()
//
//	translates lattice to world coordinates
//

void RenderPG::Unimatrix(LatticePG* ref)
{
	Entry<MeshPG>*		meshi;
	meshi = (Entry<MeshPG>*)ref->meshes;
	
	while(meshi != nil)
	{
		Entry<MeshPG>*	nextMesh;
		MeshPG*			theMesh;
		nextMesh = meshi->next;
		theMesh = meshi->content;
		
		short	xOrigin = theMesh->origin->wx;
		short	yOrigin = theMesh->origin->wy;
		short	zOrigin = theMesh->origin->wz;
		
		Entry<VertexPG>*	current;
		current = theMesh->vertices;
		while(current != nil)
		{
			VertexPG*			vertex;
			Entry<VertexPG>*	nextEntry;
			vertex = current->content;
			nextEntry = current->next;
			vertex->wx = vertex->tx + xOrigin;
			vertex->wy = vertex->ty + yOrigin;
			vertex->wz = vertex->tz + zOrigin;
			
			current = nextEntry;
		}
		
		meshi = nextMesh;
	}
}

//
//	RenderPG::Project()
//

void RenderPG::Project(LatticePG* ref)
{
	VertexPG*			vertex;
	short			horzCenter = width / 2;
	short			vertCenter = height / 2;
	
	//	do the lattice origin project
	vertex = ref->origin;
	vertex->sx = (vertex->wx * kProjectConstant);
	vertex->sy = (vertex->wy * kProjectConstant);
			
	if(vertex->wz != 0)
	{
		vertex->sx /= vertex->wz;
		vertex->sy /= vertex->wz;
	}
	
	vertex->sx += horzCenter;
	vertex->sy += vertCenter;
	
	//	now cycle thru the meshes
	Entry<MeshPG>*			meshi;
	meshi = (Entry<MeshPG>*)ref->meshes;
	while(meshi != nil)
	{
		Entry<MeshPG>*		nextMesh;
		MeshPG*				theMesh;
		nextMesh = meshi->next;
		theMesh = meshi->content;
		
		//	do the mesh offset project
		vertex = theMesh->origin;
		vertex->sx = (vertex->wx * kProjectConstant);
		vertex->sy = (vertex->wy * kProjectConstant);
				
		if(vertex->wz != 0)
		{
			vertex->sx /= vertex->wz;
			vertex->sy /= vertex->wz;
		}
		
		vertex->sx += horzCenter;
		vertex->sy += vertCenter;
		
		//	now project the vertices
		Entry<TrianglePG>*		current;
		current = theMesh->triangles;
		while(current != nil)
		{
			Entry<TrianglePG>*		nextEntry;
			TrianglePG*				triangle;
			triangle = current->content;
			nextEntry = current->next;
			
			for(short i = 0; i < triangle->clipCount; i++)
			{
				vertex = &triangle->clipped[i];
				
				vertex->x = (vertex->x * kProjectConstant);
				vertex->y = (vertex->y * kProjectConstant);
				
				if(vertex->z != 0)
				{
					vertex->x /= vertex->z;
					vertex->y /= vertex->z;
				}
				
				vertex->x += horzCenter;
				vertex->y += vertCenter;
			}
			current = nextEntry;
		}
		meshi = nextMesh;
	}
}

//
//	RenderPG::Extents()
//

void RenderPG::Extents(void)
{
	Entry<TrianglePG>*		current;
	short					xMin, xMax;
	short					yMin, yMax;
	short					zMin, zMax;
	short					edgeCount;
	
	current = (Entry<TrianglePG>*)triangles;
	while(current != nil)
	{
		TrianglePG*			tri;
		tri = current->content;
		edgeCount = tri->edgeCount - 1;
		
		xMin = xMax = tri->clipped[0].x;
		yMin = yMax = tri->clipped[0].y;
		zMin = zMax = tri->clipped[0].z;
		
		for(short i = 1; i < edgeCount; i++)
		{
			if(xMin > tri->clipped[i].x)
				xMin = tri->clipped[i].x;
				
			if(xMax < tri->clipped[i].x)
				xMax = tri->clipped[i].x;
				
			if(yMin > tri->clipped[i].y)
				yMin = tri->clipped[i].y;
				
			if(yMax < tri->clipped[i].y)
				yMax = tri->clipped[i].y;
				
			if(zMin > tri->clipped[i].z)
				zMin = tri->clipped[i].z;
				
			if(zMax < tri->clipped[i].z)
				zMax = tri->clipped[i].z;
		}
		
		tri->xMin = xMin;
		tri->yMin = yMin;
		tri->zMin = zMin;
		tri->xMax = xMax;
		tri->yMax = yMax;
		tri->zMax = zMax;
		
		current = current->next;
	}
}

//
//	RenderPG::Swap()
//

void RenderPG::Swap(Entry<TrianglePG>* left, Entry<TrianglePG>* right)
{
	TrianglePG*	temp;
	
	temp = right->content;
	right->content = left->content;
	left->content = temp;
}

//
//	RenderPG::ZSort()
//

void RenderPG::ZSort(void)
{
	if(triangles == nil)
		return;
		
	Entry<TrianglePG>*	left;
	Entry<TrianglePG>*	right;
	left = triangles;
	right = left->next;
	
	while(right != nil)
	{
		while(right->content->zMax > left->content->zMax)
		{
			Swap(left, right);
			
			if(left->prior == nil)
				break;
				
			right = left;
			left = left->prior;
		}
		
		left = right;
		right = right->next;
	}
}

//
//	RenderPG::Painters()
//

void RenderPG::Painters(void)
{
	if(triangles == nil)
		return;
	
	Entry<TrianglePG>*	left;
	Entry<TrianglePG>*	right;
	
	left = triangles;
	right = left->next;
	
	while(right != nil)
	{
		bool		swap = true;
		TrianglePG*	l = left->content;
		TrianglePG*	r = right->content;
		
		if(ZOverlap(l, r) == true)
		{
			if(XYOverlap(l, r) == false)
				swap = false;
			
			if(swap == true)
			{
				if(Inside(l, r) == true)
					swap = false;
			}
			
			if(swap == true)
			{
				if(Outside(l, r) == true)
					swap = false;
			}
		}
		else
		{
			swap = false;
		}
		
		if(swap == true)
		{
			Swap(left, right);
		}
			
		left = right;
		right = right->next;
	}
}

//
//	RenderPG::Backface()
//

bool RenderPG::Backface(TrianglePG* triangle)
{
	if(triangle->edgeCount < 3)
		return false;
	
	VertexPG*	v0;
	VertexPG*	v1;
	VertexPG*	v2;
	
	long		x1, x2, x3;
	long		y1, y2, y3;
	long		z1, z2, z3;
	long		c;
	
	v0 = triangle->vertices[0];
	v1 = triangle->vertices[1];
	v2 = triangle->vertices[2];
	
	x1 = v0->wx;	x2 = v1->wx;	x3 = v2->wx;
	y1 = v0->wy;	y2 = v1->wy;	y3 = v2->wy;
	z1 = v0->wz;	z2 = v1->wz;	z3 = v2->wz;
	
	c = (	x3 * ((z1 * y2) - (y1 * z2)) +
			y3 * ((x1 * z2) - (z1 * x2)) +
			z3 * ((y1 * x2) - (x1 * y2)));
			
	if(c > 0)
		return false;
	
	return true;
}	

//
//	RenderPG::ZOverlap()
//

bool RenderPG::ZOverlap(TrianglePG* left, TrianglePG* right)
{
	bool swap = true;
	
	if(right->zMin > left->zMax)
	{
	//	if(left->zMin > right->zMin)
			swap = false;
	} 
	
	if(left->zMax > right->zMin)
	{
	//	if(right->zMax > left->zMax)
			swap = false;
	}
	
	return swap;
}

//
//	RenderPG::XYOverlap()
//

bool RenderPG::XYOverlap(TrianglePG* left, TrianglePG* right)
{
	if((left->xMin > right->xMax) || (right->xMin > left->xMax))
		return false;

	if((left->yMin > right->yMax) || (right->yMin > left->yMax))
		return false;
	
	return true;
}

//
//	RenderPG::Inside()
//

bool RenderPG::Inside(TrianglePG* left, TrianglePG* right)
{
	long	x1, x2, x3;
	long	y1, y2, y3;
	long	z1, z2, z3;
	long	a,b,c,d;
	
	x1 = right->vertices[0]->wx;
	y1 = right->vertices[0]->wy;
	z1 = right->vertices[0]->wz;
	x2 = right->vertices[1]->wx;
	y2 = right->vertices[1]->wy;
	z2 = right->vertices[1]->wz;
	x3 = right->vertices[2]->wx;
	y3 = right->vertices[2]->wy;
	z3 = right->vertices[2]->wz;
	
	a = y1 * (z2 - z3) + y2 * (z3 - z1) + y3 * (z1 - z2);
	b = z1 * (x2 - x3) + z2 * (x3 - x1) + z3 * (x1 - x2);
	c = x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2);
	d = -x1 * (y2 * z3 - y3 * z2) - 
		x2 * (y3 * z1 - y1 * z3) - 
		x3 * (y1 * z2 - y2 * z1);
	
	for(short i = 0; i < 3; i++)
	{
		long	equationResult;
		equationResult = ((a * left->vertices[i]->wx) + 
						  (b * left->vertices[i]->wy) +
						  (c * left->vertices[i]->wz) + d);
		
		if (equationResult < 0)
			return false;
	}
	
	return true;
}

//
//	RenderPG::Outside()
//

bool RenderPG::Outside(TrianglePG* left, TrianglePG* right)
{
	long	x1, x2, x3;
	long	y1, y2, y3;
	long	z1, z2, z3;
	long	a,b,c,d;
	
	x1 = left->vertices[0]->wx;
	y1 = left->vertices[0]->wy;
	z1 = left->vertices[0]->wz;
	x2 = left->vertices[1]->wx;
	y2 = left->vertices[1]->wy;
	z2 = left->vertices[1]->wz;
	x3 = left->vertices[2]->wx;
	y3 = left->vertices[2]->wy;
	z3 = left->vertices[2]->wz;
	
	a = y1 * (z2 - z3) + y2 * (z3 - z1) + y3 * (z1 - z2);
	b = z1 * (x2 - x3) + z2 * (x3 - x1) + z3 * (x1 - x2);
	c = x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2);
	d = -x1 * (y2 * z3 - y3 * z2) - 
		x2 * (y3 * z1 - y1 * z3) - 
		x3 * (y1 * z2 - y2 * z1);
	
	for(short i = 0; i < 3; i++)
	{
		long	equationResult;
		equationResult = ((a * right->vertices[i]->wx) + 
						  (b * right->vertices[i]->wy) +
						  (c * right->vertices[i]->wz) + d);
		
		if (equationResult > 0)
			return false;
	}
	
	return true;
}

//
//	RenderPG::ZClip()
//

void RenderPG::ZClip(LatticePG* latti)
{
	Entry<TrianglePG>*	current;
	Entry<MeshPG>*		entry;
	MeshPG*				meshi;
	TrianglePG*			triangle;
	long				zMin;
	
	entry = (Entry<MeshPG>*)latti->meshes;
	while(entry != nil)
	{
		meshi = entry->content;
		current = (Entry<TrianglePG>*)meshi->triangles;
		zMin = 80;
	
		while (current != nil)
		{
			VertexPG*		clipList;
			TexturePG*		texture;
			short			v1,v2;
			VertexPG		*pv1, *pv2;
			float			t;
			short			numOfClippedVerts;
			short			textureX1, textureX2, textureY1, textureY2;
			bool			textMapped = false;
			
			triangle = current->content;
			clipList = triangle->clipped;
			texture = triangle->texture;
			numOfClippedVerts = 0;
			v1 = triangle->edgeCount-1;
			
			if(texture->mode == TexturePG::kTexture)
				textMapped = true;
			
			// begin looping through all the edges of the triangle
			for(v2 = 0; v2 < triangle->edgeCount; v2++)
			{
				pv1 = triangle->vertices[v1];
				pv2 = triangle->vertices[v2];
				
				if(textMapped == true)
				{
					textureX1 = texture->vertices[v1]->sx;
					textureY1 = texture->vertices[v1]->sy;
					textureX2 = texture->vertices[v2]->sx;
					textureY2 = texture->vertices[v2]->sy;
				}
					
				if((pv1->wz >= zMin) && (pv2->wz >= zMin))
				{
					clipList->x = pv2->wx;
					clipList->y = pv2->wy;
					clipList->z = pv2->wz;
					
					if(textMapped == true)
					{
						clipList->tx = textureX2;
						clipList->ty = textureY2;
					}
					
					clipList++;
					numOfClippedVerts++;
				} 
				
				else if((pv1->wz < zMin) && (pv2->wz < zMin))
				{
					// Edge is entirely past front; do nothing
				} 
				
				else if((pv1->wz >= zMin) && (pv2->wz < zMin))
				{
					// Edge is leaving view volume 
					t = (float)(zMin - pv1->wz) / 
						(float)(pv2->wz - pv1->wz);
					clipList->x = (short int)(pv1->wx + (pv2->wx - pv1->wx) * t);
					clipList->y = (short int)(pv1->wy + (pv2->wy - pv1->wy) * t);
					clipList->z = zMin;
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						clipList->tx = (short int)(((textureX2 - textureX1) * t) +
										textureX1);
						clipList->ty = (short int)(((textureY2 - textureY1) * t) +
										textureY1);
					}
						
					clipList++;
					numOfClippedVerts++;
				} 
				
				else if((pv1->wz < zMin) && (pv2->wz >= zMin))
				{
					// Edge is entering view volume 
					t = (float)(zMin - pv1->wz) / 
						(float)(pv2->wz - pv1->wz);
					
					clipList->x = (short int)(pv1->wx + (pv2->wx - pv1->wx) * t);
					clipList->y = (short int)(pv1->wy + (pv2->wy - pv1->wy) * t);
					clipList->z = zMin;
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						clipList->tx = (short int)(((textureX2 - textureX1) * t) +
										 textureX1);
						clipList->ty = (short int)(((textureY2 - textureY1) * t) +
										 textureY1);
					}
					
					clipList++;
					
					// Copy the vertex coordinates
					clipList->x = pv2->wx;
					clipList->y = pv2->wy;
					clipList->z = pv2->wz;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{				
						clipList->tx = textureX2;
						clipList->ty = textureY2;
					}
					
					clipList++;
					numOfClippedVerts += 2;
				}
				
				v1 = v2;
			}
			
			triangle->clipCount = numOfClippedVerts;
			current = current->next;
		}
		entry = entry->next;
	}
}

//
//	RenderPG::XYClip()
//

void RenderPG::XYClip(void)
{
	Entry<TrianglePG>*	current;
	TrianglePG*			triangle;
	long				xMin, xMax, yMin, yMax;
	
	current = triangles;
	xMin = 0;
	xMax = width - 1;
	yMin = 0;
	yMax = height - 1;
	
	while(current != nil)
	{
		VertexPG*		clipList;
		VertexPG		tempClipList[20];
		short			clipRef;
		short			v1,v2;
		float			t;
		float			texturePos;
		long			pt0y, pt1y;
		long			pt0x, pt1x;
		bool			textMapped = false;
		
		triangle = current->content;
		clipList = triangle->clipped;
		if(triangle->texture->mode == TexturePG::kTexture)
			textMapped = true;
		
		if(triangle->clipCount > 2)
		{
			// Clip against the left side and put it into the temp list
			clipRef = 0;
			v1 = triangle->clipCount-1;
			
			for(v2 = 0; v2 < triangle->clipCount; v2++)
			{
				if((clipList[v1].x >= xMin) && (clipList[v2].x >= xMin))
				{
					// Edge isn't off right side of viewport, copy it over
					tempClipList[clipRef].x = clipList[v2].x;
					tempClipList[clipRef].y = clipList[v2].y;
					tempClipList[clipRef].z = clipList[v2].z;
					
					// Copy over the texture vertices
					if(textMapped == true)
					{
						tempClipList[clipRef].tx = clipList[v2].tx;
						tempClipList[clipRef].ty = clipList[v2].ty;
					}
					clipRef++;
				}
				else if((clipList[v1].x < xMin) && (clipList[v2].x < xMin)) 
				{
					// Edge is entirely off viewport; do nothing
				}
				else if((clipList[v1].x >= xMin) && (clipList[v2].x < xMin))
				{
					// Edge is leaving viewport
					t = (float)(xMin - clipList[v1].x) / 
						(float)(clipList[v2].x - clipList[v1].x);
						
					tempClipList[clipRef].x = xMin;
					tempClipList[clipRef].y = (short int)(clipList[v1].y + 
												(clipList[v2].y - clipList[v1].y) * t);
					tempClipList[clipRef].z = (short int)(clipList[v1].z + 
												(clipList[v2].z - clipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0x = clipList[v1].x * clipList[v1].z;
						pt1x = clipList[v2].x * clipList[v2].z;
						
						texturePos = (((float)(pt0x)) /
									  (float)(pt0x - pt1x));
					
						tempClipList[clipRef].tx = (short int)(((clipList[v2].tx - clipList[v1].tx) * 
													 texturePos) + clipList[v1].tx);
						tempClipList[clipRef].ty = (short int)(((clipList[v2].ty - clipList[v1].ty) * 
												     texturePos) + clipList[v1].ty);
					}
					
					clipRef++;
				}
				else if((clipList[v1].x < xMin) && (clipList[v2].x >= xMin))
				{
					// Edge is entering viewport
					t = (float)(xMin - clipList[v1].x) / 
						(float)(clipList[v2].x - clipList[v1].x);
					tempClipList[clipRef].x = xMin;
					tempClipList[clipRef].y = (short int)(clipList[v1].y + 
												(clipList[v2].y - clipList[v1].y) * t);
					tempClipList[clipRef].z = (short int)(clipList[v1].z + 
												(clipList[v2].z - clipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0x = clipList[v1].x * clipList[v1].z;
						pt1x = clipList[v2].x * clipList[v2].z;
						
						texturePos = (((float)(pt0x)) /
									   (float)(pt0x - pt1x));
							
						tempClipList[clipRef].tx = (short int)(((clipList[v2].tx - clipList[v1].tx) * 
													 texturePos) + clipList[v1].tx);
						tempClipList[clipRef].ty = (short int)(((clipList[v2].ty - clipList[v1].ty) * 
												     texturePos) + clipList[v1].ty);
					}
					
					clipRef++;
					
					// Copy the next vertex over
					tempClipList[clipRef].x = clipList[v2].x;
					tempClipList[clipRef].y = clipList[v2].y;
					tempClipList[clipRef].z = clipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{	
						tempClipList[clipRef].tx = clipList[v2].tx;
						tempClipList[clipRef].ty = clipList[v2].ty;
					}
					
					clipRef++;
				}
				
				v1 = v2;
			}
			triangle->clipCount = clipRef;
		}
		
		if(triangle->clipCount > 2)
		{
			// Clip against the right edge of the viewport
			
			clipRef = 0;
			v1 = triangle->clipCount-1;
			
			for(v2 = 0; v2 < triangle->clipCount; v2++)
			{
				if((tempClipList[v1].x <= xMax) && (tempClipList[v2].x <= xMax))
				{
					// Edge isn't off right side of viewport; copy it over
					clipList[clipRef].x = tempClipList[v2].x;
					clipList[clipRef].y = tempClipList[v2].y;
					clipList[clipRef].z = tempClipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						clipList[clipRef].tx = tempClipList[v2].tx;
						clipList[clipRef].ty = tempClipList[v2].ty;
					}
					
					clipRef++;
				}
				else if((tempClipList[v1].x > xMax) && (tempClipList[v2].x > xMax))
				{
					// Edge is entirely off viewport; do nothing
				}
				else if((tempClipList[v1].x <= xMax) && (tempClipList[v2].x > xMax))
				{
					// Edge is leaving viewport
					t = (float)(xMax - tempClipList[v1].x) / 
						(float)(tempClipList[v2].x - tempClipList[v1].x);
					clipList[clipRef].x = xMax;
					clipList[clipRef].y = (short int)(tempClipList[v1].y + 
											(tempClipList[v2].y - tempClipList[v1].y) * t);
					clipList[clipRef].z = (short int)(tempClipList[v1].z + 
											(tempClipList[v2].z - tempClipList[v1].z) * t);
					
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0x = (xMax - tempClipList[v1].x) * tempClipList[v1].z;
						pt1x = (xMax - tempClipList[v2].x) * tempClipList[v2].z;
						
						texturePos = (((float)(pt0x)) /
									  (float)(pt0x - pt1x));
					
						clipList[clipRef].tx = (short int)(((tempClipList[v2].tx - tempClipList[v1].tx) * 
												 texturePos) + tempClipList[v1].tx);
						clipList[clipRef].ty = (short int)(((tempClipList[v2].ty - tempClipList[v1].ty) * 
											     texturePos) + tempClipList[v1].ty);
					}
					clipRef++;
				}
				else if ((tempClipList[v1].x > xMax) && (tempClipList[v2].x <= xMax))
				{
					// Edge is entering viewport
					t = (float)(xMax - tempClipList[v1].x) / 
						(float)(tempClipList[v2].x - tempClipList[v1].x);
					clipList[clipRef].x = xMax;
					clipList[clipRef].y = (short int)(tempClipList[v1].y + 
											(tempClipList[v2].y - tempClipList[v1].y) * t);
					clipList[clipRef].z = (short int)(tempClipList[v1].z + 
											(tempClipList[v2].z - tempClipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0x = (xMax - tempClipList[v1].x) * tempClipList[v1].z;
						pt1x = (xMax - tempClipList[v2].x) * tempClipList[v2].z;
						
						texturePos = (((float)(pt0x)) /
									  (float)(pt0x - pt1x));
						
						clipList[clipRef].tx = (short int)(((tempClipList[v2].tx - tempClipList[v1].tx) * 
												 texturePos) + tempClipList[v1].tx);
						clipList[clipRef].ty = (short int)(((tempClipList[v2].ty - tempClipList[v1].ty) * 
											     texturePos) + tempClipList[v1].ty);
					}

					clipRef++;
					
					clipList[clipRef].x = tempClipList[v2].x;
					clipList[clipRef].y = tempClipList[v2].y;
					clipList[clipRef].z = tempClipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						clipList[clipRef].tx = tempClipList[v2].tx;
						clipList[clipRef].ty = tempClipList[v2].ty;
					}
					
					clipRef++;
				}
				
				v1 = v2;
			}
			triangle->clipCount = clipRef;
		}		

		if (triangle->clipCount > 2)
		{
			// Clip against the top side of the viewport
			clipRef = 0;
			v1 = triangle->clipCount-1;
			
			for(v2 = 0; v2 < triangle->clipCount; v2++)
			{
				if((clipList[v1].y >= yMin) && (clipList[v2].y >= yMin))
				{
					// Edge isn't off top side of viewport; copy it over
					tempClipList[clipRef].x = clipList[v2].x;
					tempClipList[clipRef].y = clipList[v2].y;
					tempClipList[clipRef].z = clipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						tempClipList[clipRef].tx = clipList[v2].tx;
						tempClipList[clipRef].ty = clipList[v2].ty;
					}
					
					clipRef++;
				}
				else if ((clipList[v1].y < yMin) && (clipList[v2].y < yMin))
				{
					// Edge is entirely off viewport; do nothing
				}
				else if((clipList[v1].y >= yMin) && (clipList[v2].y < yMin))
				{
					// Edge is leaving viewport
					t = (float)(yMin - clipList[v1].y) / 
						(float)(clipList[v2].y - clipList[v1].y);
					tempClipList[clipRef].x = (short int)(clipList[v1].x + 
												(clipList[v2].x - clipList[v1].x) * t);
					
					tempClipList[clipRef].y = yMin;
					tempClipList[clipRef].z = (short int)(clipList[v1].z + 
												(clipList[v2].z - clipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0y = clipList[v1].y;
						pt1y = clipList[v2].y;
						
						texturePos = (((float)(pt0y)) /
									   (float)(pt0y - pt1y));
							
						tempClipList[clipRef].tx = (short int)(((clipList[v2].tx - clipList[v1].tx) * 
													 texturePos) + clipList[v1].tx);
						tempClipList[clipRef].ty = (short int)(((clipList[v2].ty - clipList[v1].ty) * 
												     texturePos) + clipList[v1].ty);
					}
					
					clipRef++;
				}
				else if((clipList[v1].y < yMin) && (clipList[v2].y >= yMin))
				{
					// Edge is entering viewport
					t = (float)(yMin - clipList[v1].y) / 
						(float)(clipList[v2].y - clipList[v1].y);
					tempClipList[clipRef].x = (short int)(clipList[v1].x + 
												(clipList[v2].x - clipList[v1].x) * t);
					
					tempClipList[clipRef].y = yMin;
					tempClipList[clipRef].z = (short int)(clipList[v1].z + 
												(clipList[v2].z - clipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0y = clipList[v1].y;
						pt1y = clipList[v2].y;
						
						texturePos = (((float)(pt0y)) /
									   (float)(pt0y - pt1y));
					
						tempClipList[clipRef].tx = (short int)(((clipList[v2].tx - clipList[v1].tx) * 
													 texturePos) + clipList[v1].tx);
						tempClipList[clipRef].ty = (short int)(((clipList[v2].ty - clipList[v1].ty) * 
												     texturePos) + clipList[v1].ty);
					}
					
					clipRef++;
					
					tempClipList[clipRef].x = clipList[v2].x;
					tempClipList[clipRef].y = clipList[v2].y;
					tempClipList[clipRef].z = clipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						tempClipList[clipRef].tx = clipList[v2].tx;
						tempClipList[clipRef].ty = clipList[v2].ty;
					}
					
					clipRef++;
				}
				v1 = v2;
			}
			triangle->clipCount = clipRef;
		}		
		
		if (triangle->clipCount > 2)
		{
			// Clip against the bottom edge of the viewport */
			clipRef = 0;
			v1 = triangle->clipCount-1;
			
			for(v2 = 0; v2 < triangle->clipCount; v2++)
			{
				if((tempClipList[v1].y <= yMax) && (tempClipList[v2].y <= yMax))
				{
					// Edge isn't off edge side of viewport; copy it over
					clipList[clipRef].x = tempClipList[v2].x;
					clipList[clipRef].y = tempClipList[v2].y;
					clipList[clipRef].z = tempClipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						clipList[clipRef].tx = tempClipList[v2].tx;
						clipList[clipRef].ty = tempClipList[v2].ty;
					}
					
					clipRef++;
				}
				
				else if((tempClipList[v1].y > yMax) && (tempClipList[v2].y > yMax))
				{
					// Edge is entirely off viewport; do nothing
				}
				
				else if((tempClipList[v1].y <= yMax) && (tempClipList[v2].y > yMax))
				{
					// Edge is leaving viewport
					t = (float)(yMax - tempClipList[v1].y) / 
						(float)(tempClipList[v2].y - tempClipList[v1].y);
					clipList[clipRef].x = (short int)(tempClipList[v1].x + 
												(tempClipList[v2].x - tempClipList[v1].x) * t);
					
					clipList[clipRef].y = yMax;
					clipList[clipRef].z = (short int)(tempClipList[v1].z + 
												(tempClipList[v2].z - tempClipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0y = (yMax - tempClipList[v1].y);
						pt1y = (yMax - tempClipList[v2].y);
						
						texturePos = (((float)(pt0y)) /
									   (float)(pt0y - pt1y));
						
						clipList[clipRef].tx = (short int)(((tempClipList[v2].tx - tempClipList[v1].tx) * 
												 texturePos) + tempClipList[v1].tx);
						clipList[clipRef].ty = (short int)(((tempClipList[v2].ty - tempClipList[v1].ty) * 
											     texturePos) + tempClipList[v1].ty);
					}

					clipRef++;
				}
				
				else if((tempClipList[v1].y > yMax) && (tempClipList[v2].y <= yMax))
				{
					// Edge is entering viewport
					t = (float)(yMax - tempClipList[v1].y) / 
						(float)(tempClipList[v2].y - tempClipList[v1].y);
					clipList[clipRef].x = (short int)(tempClipList[v1].x + 
												(tempClipList[v2].x - tempClipList[v1].x) * t);
					
					clipList[clipRef].y = yMax;
					clipList[clipRef].z = (short int)(tempClipList[v1].z + 
												(tempClipList[v2].z - tempClipList[v1].z) * t);
					
					// Calculate the texture coordinates
					if(textMapped == true)
					{
						pt0y = (yMax - tempClipList[v1].y);
						pt1y = (yMax - tempClipList[v2].y);
						
						texturePos = (((float)(pt0y)) /
									   (float)(pt0y - pt1y));
						
						clipList[clipRef].tx = (short int)(((tempClipList[v2].tx - tempClipList[v1].tx) * 
												 texturePos) + tempClipList[v1].tx);
						clipList[clipRef].ty = (short int)(((tempClipList[v2].ty - tempClipList[v1].ty) * 
											     texturePos) + tempClipList[v1].ty);
					}

					clipRef++;
					
					clipList[clipRef].x = tempClipList[v2].x;
					clipList[clipRef].y = tempClipList[v2].y;
					clipList[clipRef].z = tempClipList[v2].z;
					
					// Copy the texture coordinates
					if(textMapped == true)
					{
						clipList[clipRef].tx = tempClipList[v2].tx;
						clipList[clipRef].ty = tempClipList[v2].ty;
					}
					
					clipRef++;
				}
				
				v1 = v2;
			}
			triangle->clipCount = clipRef;
		}
		current = current->next;
	}

}

//#pragma mark -

//
//	RenderPG::DrawTriangles()
//

void RenderPG::DrawTriangles(void)
{
	Entry<TrianglePG>* 	current;
	
	current = triangles;
	while(current != nil)
	{
		Entry<TrianglePG>*	next;
		TrianglePG*			tri;
		next = current->next;
		tri = current->content;
		
		RenderTriangle(tri);
		
		if(next == nil)
			return;
		
		current = next;
	}
}

//
//	RenderPG::RenderTriangle()
//

void RenderPG::RenderTriangle(TrianglePG* tri)
{
	if(!Backface(tri))
	{
		switch(tri->texture->mode)
		{
			case TexturePG::kInvisable:
			//	don't draw
			break;
			
			case TexturePG::kWireframe:
			RenderTriangleWire(tri);
			break;
			
			case TexturePG::kPolygon:
			RenderTrianglePoly(tri);
			break;
			
			case TexturePG::kTexture:
			RenderTriangleText(tri);
			break;
		}
	}
}

//
//	RenderPG::RenderTriangleWire()
//

void RenderPG::RenderTriangleWire(TrianglePG* triangle)
{
	// find the top triangle point
	short 		i;
	VertexPG*	vertexList = triangle->clipped;
	short 		topIndex;		//	index of the top vertex
	short		yMin;			//	top y-Coord
	short		edgeCount = (triangle->clipCount) -1;
	short		totalEdges = edgeCount;
	
	topIndex = 0;
	yMin = vertexList[topIndex].y;
	for(i = 1; i <= totalEdges; i++)
	{
		short yTemp = vertexList[i].y;
		
		if(yMin > yTemp)
		{
			yMin = yTemp;
			topIndex = i;
		}
	}
	
	//	get the starting values
	short	xStart1, xEnd1;
	short	xStart2, xEnd2;
	short	yStart1, yEnd1;
	short	yStart2, yEnd2;
	short	startVert1;
	short	startVert2;
	short	endVert1;
	short	endVert2;
	
	startVert1 = topIndex;
	startVert2 = topIndex;
	
	xStart1 = vertexList[startVert1].x;
	yStart1 = vertexList[startVert1].y;
	xStart2 = vertexList[startVert2].x;
	yStart2 = vertexList[startVert2].y;
	
	endVert1 = topIndex - 1;
	if(endVert1 < 0)
		endVert1 = totalEdges;
		
	endVert2 = topIndex + 1;
	if(endVert2 > totalEdges)
		endVert2 = 0;
	
	xEnd1 = vertexList[endVert1].x;
	yEnd1 = vertexList[endVert1].y;
	xEnd2 = vertexList[endVert2].x;
	yEnd2 = vertexList[endVert2].y;
	
	//	calculate the drawing values
	short	dx1, dx2;
	short	dy1, dy2;
	long	leftDelta;
	long	rightDelta;
	long	xStart, xEnd;
	short	yStart, yEnd;
	
	dx1 = xEnd1 - xStart1;
	dx2 = xEnd2 - xStart2;
	dy1 = yEnd1 - yStart1;
	dy2 = yEnd2 - yStart2;
	
	if(dy1 != 0)
		leftDelta = (dx1 << 8) / dy1;
		
	if(dy2 != 0)
		rightDelta = (dx2 << 8) / dy2;
	
	xStart = (xStart1 << 8);
	xEnd = (xStart2 << 8);
	
	yStart = yStart1;
	yEnd = yStart2;
	
	// the drawing loop
	while(edgeCount > 0)
	{
		if(dy1 && dy2)
		{
			do
			{
				unsigned char	color = triangle->texture->color;
				unsigned long	index;
				short			startLeft = (xStart >> 8);
				short			startRight = (xEnd >> 8);
				short			endLeft;
				short			endRight;
				short 			x;
				
				//	incrament the x vals
				//	to determine our stopping location
				xStart += leftDelta;
				xEnd += rightDelta;
				
				endLeft = (xStart >> 8);
				endRight  = (xEnd >> 8);
				
				if(startLeft > endLeft)
				{
					i = startLeft;
					startLeft = endLeft;
					endLeft = i;
				}
				
				if(startRight > endRight)
				{
					i = startRight;
					startRight = endRight;
					endRight = i;
				}
				
				//	swap values if necessary
				if(startLeft > startRight)
				{
					i = startLeft;
					startLeft = startRight;
					startRight = i;
					
					i = endLeft;
					endLeft = endRight;
					endRight = i;
				}
				
				// scale the left side
				for(x = startLeft; x <= endLeft; x++)
				{
					index = (yStart * rowBytes) + x;
					baseAddress[index] = color;
				}
					
				// scale the right side
				for(x = startRight; x <= endRight; x++)
				{
					index = (yStart * rowBytes) + x;
					baseAddress[index] = color;
				}

				//	increment our scanline values
				yStart++;
				yEnd++;
			}
			while((yStart < yEnd1) && (yEnd < yEnd2));
		}
		
		//	handle line ending
		
		if((yStart == yEnd1) || (!dy1))
		{
			startVert1 = endVert1;		
			
			//	left line ended
			xStart1 = xEnd1;
			yStart1 = yEnd1;
			
			endVert1 = startVert1 - 1;
			if(endVert1 < 0)
				endVert1 = totalEdges;
				
			xEnd1 = vertexList[endVert1].x;
			yEnd1 = vertexList[endVert1].y;
			
			dx1 = xEnd1 - xStart1;
			dy1 = yEnd1 - yStart1;
			if(dy1 != 0)
				leftDelta = (dx1 << 8) / dy1;
				
			xStart = xStart1 << 8;
			edgeCount--;
		}
		
		if((yEnd == yEnd2) || (!dy2))
		{
			startVert2 = endVert2;
			
			//	right line ended
			xStart2 = xEnd2;
			yStart2 = yEnd2;
			
			endVert2 = startVert2 + 1;
			if(endVert2 > totalEdges)
				endVert2 = 0;
			
			xEnd2 = vertexList[endVert2].x;
			yEnd2 = vertexList[endVert2].y;
			
			dx2 = xEnd2 - xStart2;
			dy2 = yEnd2 - yStart2;
			if(dy2 != 0)
				rightDelta = (dx2 << 8) / dy2;
				
			xEnd = xStart2 << 8;
			edgeCount--;
		}
	}
}

//
//	RenderPG::RenderTrianglePoly()
//

void RenderPG::RenderTrianglePoly(TrianglePG* triangle)
{
	// find the top triangle point
	short 		i;
	VertexPG*	vertexList = triangle->clipped;
	short 		topIndex;		//	index of the top vertex
	short		yMin;			//	top y-Coord
	short		edgeCount = (triangle->clipCount) -1;
	short		totalEdges = edgeCount;
	
	topIndex = 0;
	yMin = vertexList[topIndex].y;
	for(i = 1; i <= totalEdges; i++)
	{
		short yTemp = vertexList[i].y;
		
		if(yMin > yTemp)
		{
			yMin = yTemp;
			topIndex = i;
		}
	}
	
	//	get the starting values
	short	xStart1, xEnd1;
	short	xStart2, xEnd2;
	short	yStart1, yEnd1;
	short	yStart2, yEnd2;
	short	startVert1;
	short	startVert2;
	short	endVert1;
	short	endVert2;
	
	startVert1 = topIndex;
	startVert2 = topIndex;
	
	xStart1 = vertexList[startVert1].x;
	yStart1 = vertexList[startVert1].y;
	xStart2 = vertexList[startVert2].x;
	yStart2 = vertexList[startVert2].y;
	
	endVert1 = topIndex - 1;
	if(endVert1 < 0)
		endVert1 = totalEdges;
		
	endVert2 = topIndex + 1;
	if(endVert2 > totalEdges)
		endVert2 = 0;
	
	xEnd1 = vertexList[endVert1].x;
	yEnd1 = vertexList[endVert1].y;
	xEnd2 = vertexList[endVert2].x;
	yEnd2 = vertexList[endVert2].y;
	
	//	calculate the drawing values
	short	dx1, dx2;
	short	dy1, dy2;
	FixedPG	leftDelta;
	FixedPG	rightDelta;
	FixedPG	xStart, xEnd;
	short	yStart, yEnd;
	
	dx1 = xEnd1 - xStart1;
	dx2 = xEnd2 - xStart2;
	dy1 = yEnd1 - yStart1;
	dy2 = yEnd2 - yStart2;
	
	if(dy1 != 0)
		leftDelta = (dx1 << 8) / dy1;
		
	if(dy2 != 0)
		rightDelta = (dx2 << 8) / dy2;
	
	xStart = (xStart1 << 8);
	xEnd = (xStart2 << 8);
	
	yStart = yStart1;
	yEnd = yStart2;
	
	// the drawing loop
	while(edgeCount > 0)
	{
		if(dy1 && dy2)
		{
			do
			{
				unsigned char	color = triangle->texture->color;
				short			left = (xStart >> 8);
				short			right = (xEnd >> 8);
				
				//	swap values if necessary
				if(left > right)
				{
					i = left;
					left = right;
					right = i;
				}
				
				for(short x = left; x <= right; x++)
				{
					unsigned long	index;
					index = (yStart * rowBytes) + x;
					baseAddress[index] = color;
				}

				//	increment our scanline values
				yStart++;
				yEnd++;
				
				xStart += leftDelta;
				xEnd += rightDelta;
			}
			while((yStart < yEnd1) && (yEnd < yEnd2));
		}
		
		//	handle line ending
		
		if((yStart == yEnd1) || (!dy1))
		{
			startVert1 = endVert1;		
			
			//	left line ended
			xStart1 = xEnd1;
			yStart1 = yEnd1;
			
			endVert1 = startVert1 - 1;
			if(endVert1 < 0)
				endVert1 = totalEdges;
				
			xEnd1 = vertexList[endVert1].x;
			yEnd1 = vertexList[endVert1].y;
			
			dx1 = xEnd1 - xStart1;
			dy1 = yEnd1 - yStart1;
			if(dy1 != 0)
				leftDelta = (dx1 << 8) / dy1;
				
			xStart = xStart1 << 8;
			edgeCount--;
		}
		
		if((yEnd == yEnd2) || (!dy2))
		{
			startVert2 = endVert2;
			
			//	right line ended
			xStart2 = xEnd2;
			yStart2 = yEnd2;
			
			endVert2 = startVert2 + 1;
			if(endVert2 > totalEdges)
				endVert2 = 0;
			
			xEnd2 = vertexList[endVert2].x;
			yEnd2 = vertexList[endVert2].y;
			
			dx2 = xEnd2 - xStart2;
			dy2 = yEnd2 - yStart2;
			if(dy2 != 0)
				rightDelta = (dx2 << 8) / dy2;
				
			xEnd = xStart2 << 8;
			edgeCount--;
		}
	}
}

//
//	RenderPG::RenderTriangleText()
//

void RenderPG::RenderTriangleText(TrianglePG* triangle)
{
	// find the top triangle point
	short 		i;
	TexturePG*	texture = triangle->texture;
	VertexPG*	vertexList = triangle->clipped;
	short 		topIndex;		//	index of the top vertex
	short		yMin;			//	top y-Coord
	short		edgeCount = (triangle->clipCount) - 1;
	short		totalEdges = edgeCount;
	
	topIndex = 0;
	yMin = vertexList[topIndex].y;
	for(i = 1; i <= totalEdges; i++)
	{
		short yTemp = vertexList[i].y;
		
		if(yMin > yTemp)
		{
			yMin = yTemp;
			topIndex = i;
		}
	}
	
	//	get the starting values
	short	xStart1, xEnd1;
	short	xStart2, xEnd2;
	short	yStart1, yEnd1;
	short	yStart2, yEnd2;
	short	startVert1;
	short	startVert2;
	short	endVert1;
	short	endVert2;
	long	a;
	long	b;
	long	ax, ay;
	long	bx, by;
	short	txStart1, txEnd1;
	short	txStart2, txEnd2;
	short	tyStart1, tyEnd1;
	short	tyStart2, tyEnd2;
	
	startVert1 = topIndex;
	startVert2 = topIndex;
	
	xStart1 = vertexList[startVert1].x;
	yStart1 = vertexList[startVert1].y;
	xStart2 = vertexList[startVert2].x;
	yStart2 = vertexList[startVert2].y;
	
	txStart1 = vertexList[startVert1].tx;
	tyStart1 = vertexList[startVert1].ty;
	txStart2 = vertexList[startVert2].tx;
	tyStart2 = vertexList[startVert2].ty;
	
	endVert1 = (startVert1 - 1);
	if(endVert1 < 0)
		endVert1 = totalEdges;
		
	endVert2 = (startVert2 + 1);
	if(endVert2 == (totalEdges + 1))
		endVert2 = 0;
	
	xEnd1 = vertexList[endVert1].x;
	yEnd1 = vertexList[endVert1].y;
	xEnd2 = vertexList[endVert2].x;
	yEnd2 = vertexList[endVert2].y;
	
	txEnd1 = vertexList[endVert1].tx;
	tyEnd1 = vertexList[endVert1].ty;
	txEnd2 = vertexList[endVert2].tx;
	tyEnd2 = vertexList[endVert2].ty;
	
	//	calculate the drawing values
	short	dx1, dx2;
	short	dy1, dy2;
	FixedPG	leftDelta;
	FixedPG	rightDelta;
	FixedPG	xStart, xEnd;
	FixedPG	yStart, yEnd;
	short	y;
	
	dx1 = xEnd1 - xStart1;
	dx2 = xEnd2 - xStart2;
	dy1 = yEnd1 - yStart1;
	dy2 = yEnd2 - yStart2;
	
	if(dy1 != 0)
		leftDelta = (dx1 << 8) / dy1;
		
	if(dy2 != 0)
		rightDelta = (dx2 << 8) / dy2;
	
	xStart = ((long)xStart1) << 8;
	yStart = ((long)yStart1) << 8;
	xEnd = ((long)xStart2) << 8;
	yEnd = ((long)yStart2) << 8;
	y = yStart1;
	
	// the drawing loop
	while(edgeCount > 0)
	{
		if(dy1 && dy2)
		{
			do
			{
				short			left = (xStart >> 8);
				short			right = (xEnd >> 8);
				short			lineLength; 
				
				//	swap values if necessary
				if(left > right)
				{
					i = left;
					left = right;
					right = i;
				}
				
				lineLength = (xEnd - xStart) >> 8;
				
				// begin calculating our texture values
				if(lineLength > 0)
				{
					FixedPG		xTexture, yTexture;
					long		txDelta, tyDelta;
					long		txError, tyError;
					short		txDirection, tyDirection;
					
					unsigned long	dst;
					unsigned long	src;
				
					a = (yStart - (((long)yEnd1) << 8)) / dy1;
					b = (yEnd - (((long)yEnd2) << 8)) / dy2;
				
					if(a < 0)	a = -a;
					if(b < 0)	b = -b;
						
					a = 256 - a;
					b = 256 - b;
					
					ax = (txStart1 << 8) - ((txStart1 - txEnd1) * a);
					ay = (tyStart1 << 8) - ((tyStart1 - tyEnd1) * a);
					bx = (txStart2 << 8) - ((txStart2 - txEnd2) * b);
					by = (tyStart2 << 8) - ((tyStart2 - tyEnd2) * b);
					
					// calculate the values for drawing
					dst = ((rowBytes * y) + left);
					
					xTexture = ax;
					yTexture = ay;
					
					txDelta = (bx - ax) / (lineLength + 1);
					tyDelta = (by - ay) / (lineLength + 1);
					
					if(txDelta < 0)
					{
						txDirection = -1;
						txDelta = -txDelta;
					}
					else
						txDirection = 1;
						
					if(tyDelta < 0)
					{
						tyDirection = -1;
						tyDelta = -tyDelta;
					}
					else
						tyDirection = 1;
						
					txError = 0;
					tyError = 0;
					
					src = ((texture->width * (yTexture >> 8)) + (xTexture >> 8));
					
					while(lineLength >= 0)
					{
						baseAddress[dst] = texture->texture[src];
						
						txError += txDelta;
						if(txError > 255)
						{
							if(txDirection == 1)
								src += (txError >> 8);
							else
								src -= (txError >> 8);
								
							txError &= 0xFF;
						}
						
						tyError += tyDelta;
						if(tyError > 255)
						{
							if(tyDirection == 1)
								src += ((tyError >> 8) * texture->width);
							else
								src -= ((tyError >> 8) * texture->width);
							
							tyError &= 0xFF;
						}
						
						dst++;
						lineLength--;
					}
				}
				else if(lineLength < 0)
				{
					FixedPG		xTexture, yTexture;
					long		txDelta, tyDelta;
					long		txError, tyError;
					short		txDirection, tyDirection;
					
					unsigned long	dst;
					unsigned long	src;
					
					a = (yStart - (((long)yEnd1) << 8)) / dy1;
					b = (yEnd - (((long)yEnd2) << 8)) / dy2;
					
					if(a < 0)	a = -a;
					if(b < 0)	b = -b;
						
					a = 256 - a;
					b = 256 - b;
					
					ax = (txStart1 << 8) - ((txStart1 - txEnd1) * a);
					ay = (tyStart1 << 8) - ((tyStart1 - tyEnd1) * a);
					bx = (txStart2 << 8) - ((txStart2 - txEnd2) * b);
					by = (tyStart2 << 8) - ((tyStart2 - tyEnd2) * b);
					
					// calculate the values for drawing
					dst = ((rowBytes * y) + right);
					
					xTexture = ax;
					yTexture = ay;
					
					txDelta = (bx - ax) / ((-lineLength) + 1);
					tyDelta = (by - ay) / ((-lineLength) + 1);
					
					if(txDelta < 0)
					{
						txDirection = -1;
						txDelta = -txDelta;
					}
					else
						txDirection = 1;
						
					if(tyDelta < 0)
					{
						tyDirection = -1;
						tyDelta = -tyDelta;
					}
					else
						tyDirection = 1;
						
					txError = 0;
					tyError = 0;
					
					src = ((texture->width * (yTexture >> 8)) + (xTexture >> 8));
					
					while(lineLength <= 0)
					{
						baseAddress[dst] = texture->texture[src];
						
						txError += txDelta;
						if(txError > 255)
						{
							if(txDirection == 1)
								src += (txError >> 8);
							else
								src -= (txError >> 8);
								
							txError &= 0xFF;
						}
						
						tyError += tyDelta;
						if(tyError > 255)
						{
							if(tyDirection == 1)
								src += ((tyError >> 8) * texture->width);
							else
								src -= ((tyError >> 8) * texture->width);
							
							tyError &= 0xFF;
						}
						
						dst--;
						lineLength++;
					}
				}

				//	increment our scanline values
				yStart += 256;
				yEnd += 256;
				y++;
				
				xStart += leftDelta;
				xEnd += rightDelta;
			}
			while(((yStart >> 8) < yEnd1) && ((yEnd >> 8) < yEnd2));
		}
		
		//	handle line ending
		
		if(((yStart >> 8) == yEnd1) || (!dy1))
		{
			startVert1 = endVert1;		
			
			//	left line ended
			xStart1 = xEnd1;
			yStart1 = yEnd1;
			txStart1 = txEnd1;
			tyStart1 = tyEnd1;
			
			endVert1 = startVert1 - 1;
			if(endVert1 < 0)
				endVert1 = totalEdges;
				
			xEnd1 = vertexList[endVert1].x;
			yEnd1 = vertexList[endVert1].y;
			txEnd1 = vertexList[endVert1].tx;
			tyEnd1 = vertexList[endVert1].ty;
			
			dx1 = xEnd1 - xStart1;
			dy1 = yEnd1 - yStart1;
			if(dy1 != 0)
				leftDelta = (dx1 << 8) / dy1;
				
			xStart = ((long)xStart1) << 8;
			yStart = ((long)yStart1) << 8;
			edgeCount--;
		}
		
		if(((yEnd >> 8) == yEnd2) || (!dy2))
		{
			startVert2 = endVert2;
			
			//	right line ended
			xStart2 = xEnd2;
			yStart2 = yEnd2;
			txStart2 = txEnd2;
			tyStart2 = tyEnd2;
			
			endVert2 = startVert2 + 1;
			if(endVert2 == (totalEdges + 1))
				endVert2 = 0;
			
			xEnd2 = vertexList[endVert2].x;
			yEnd2 = vertexList[endVert2].y;
			txEnd2 = vertexList[endVert2].tx;
			tyEnd2 = vertexList[endVert2].ty;
			
			dx2 = xEnd2 - xStart2;
			dy2 = yEnd2 - yStart2;
			if(dy2 != 0)
				rightDelta = (dx2 << 8) / dy2;
				
			xEnd = ((long)xStart2) << 8;
			yEnd = ((long)yStart2) << 8;
			edgeCount--;
		}
	}
}

//
//	RenderPG.cp
//